from representation import dessiner_monde
from initialisation import creer_monde, peupler_monde
from analyse_monde import *
from param import *

import time
from random import randint

"""
Ce programme permet de simuler ce que l'on appelle un 'système proies-prédateurs' :
il s'agit d'un monde virtuelle dans lequel évoluent des proies et des prédateurs, qui
peuvent attraper des proies pour les manger.
Ici, il s'agit d'une version très simplifiée, où les proies sont des moutons, et les
prédateurs sont des loups. Pour que les loups attrapent un mouton, il faut que le
mouton n'ait aucune échappatoire.
À chaque "tour", ce sont soit les loups soit les moutons qui se déplacent. Pour représenter
le fait que les loups sont plus rapides que les moutons, ils se déplacent en moyenne
3 fois plus souvent que les moutons.
Quand un mouton qui doit se déplacer constate qu'il ne peut pas le faire car il n'y a
pas de case libre voisine, son niveau de stress (initialement à 0) augmente. S'il
dépasse le seuil STRESS_MAX, le mouton meurt de stress. Dès qu'un mouton réussit à se
déplacer, son niveau de stress retombe à 0.
"""

def gerer_prise_mouton(monde, liste_moutons, mouton):
    """
    Cette fonction permet de déterminer si un mouton est capturé, et si c'est le cas,
    elle le supprime
    """
    ligne = mouton[INDICE_LIGNE]
    colonne = mouton[INDICE_COLONNE]
    if ligne > 0 and not est_loup(monde, ligne - 1, colonne):
        return
    if ligne < TAILLE - 1 and not est_loup(monde, ligne + 1, colonne):
        return
    if colonne > 0 and not est_loup(monde, ligne, colonne - 1):
        return 
    if colonne < TAILLE - 1 and not est_loup(monde, ligne, colonne + 1):
        return
    print("Le mouton en (", ligne, ",", colonne, ") est attrapé")
    monde[ligne][colonne] = None
    liste_moutons.remove(mouton)
    

def deplacer_animal(monde, infos, liste_moutons, animal, nouvelle_ligne, nouvelle_colonne):
    """
    Cette fonction permet de déplacer un animal (loup ou mouton) d'une case à une autre.
    Ce doit absolument être une case voisine orthogonalement.
    Si on a déplacer un loup, alors il faut vérifier si des moutons adjacents ont été
    attrapés.
    """
    ancienne_ligne = animal[INDICE_LIGNE]
    ancienne_colonne = animal[INDICE_COLONNE]
    assert abs(ancienne_ligne - nouvelle_ligne) + abs(ancienne_colonne - nouvelle_colonne) == 1, str(ancienne_ligne) + "," + str(ancienne_colonne) + " -> " + str(nouvelle_ligne) + "," + str(nouvelle_colonne)
    if monde[nouvelle_ligne][nouvelle_colonne] is not None:
        # la case cible est occupée ; on ne fait rien
        return
    monde[ancienne_ligne][ancienne_colonne] = None
    animal[INDICE_LIGNE] = nouvelle_ligne
    animal[INDICE_COLONNE] = nouvelle_colonne
    monde[nouvelle_ligne][nouvelle_colonne] = animal
    if animal[INDICE_TYPE] == TYPE_LOUP:
        moutons_adjacents = moutons_voisins(monde, nouvelle_ligne, nouvelle_colonne)
        for mouton in moutons_adjacents:
            gerer_prise_mouton(monde, liste_moutons, mouton)
    
def deplacer_un_mouton(monde, infos, liste_moutons, mouton):
    """
    Cette fonction permet de déplacer un mouton aléatoirement sur une case voisine libre.
    Si cela est impossible, le niveau de stress du mouton augmente.
    Si le niveau de stress dépasse la constante STRESS_MAX, le mouton meurt.
    """
    ligne = mouton[INDICE_LIGNE]
    colonne = mouton[INDICE_COLONNE]
    cases_possibles =  voisins_libres(monde, ligne, colonne)
    if len(cases_possibles) > 0:
        num_destination = randint(0, len(cases_possibles) - 1)
        destination = cases_possibles[num_destination]
        deplacer_animal(monde, infos, liste_moutons, mouton, destination[0], destination[1])
        mouton[INDICE_STRESS] = 0 # le stress retombe à 0
    else:
        mouton[INDICE_STRESS] += 1 #le stress augment de 1
        if mouton[INDICE_STRESS] > STRESS_MAX:
            print("Le mouton en", ligne, ",", colonne, "meurt de stress")
            monde[ligne][colonne] = None
            liste_moutons.remove(mouton)

def deplacer_les_moutons(monde, infos, liste_moutons):
    """
    Fonction faisant se déplacer tous les moutons
    """
    for mouton in liste_moutons:
        deplacer_un_mouton(monde, infos, liste_moutons, mouton)


def deplacer_un_loup_caler(monde, infos, liste_loups, liste_moutons, loup):
    """
    on insiste à gauche, puis en haut, puis à droite, puis en bas, etc.
    """
    if loup[INDICE_INFOS] is None:
        loup[INDICE_INFOS] = [0, 0, 0, 0]
    ligne = loup[INDICE_LIGNE]
    colonne = loup[INDICE_COLONNE]
    gauche_ligne, gauche_colonne = case_a_gauche(ligne, colonne)
    haut_ligne, haut_colonne = case_au_dessus(ligne, colonne)
    droite_ligne, droite_colonne = case_a_droite(ligne, colonne)
    bas_ligne, bas_colonne = case_au_dessous(ligne, colonne)
    if loup[INDICE_INFOS][0] < 20 and gauche_ligne is not None and est_vide(monde, gauche_ligne, gauche_colonne):
        deplacer_animal(monde, infos, liste_moutons, loup, gauche_ligne, gauche_colonne)
        loup[INDICE_INFOS][0] += 1
    elif loup[INDICE_INFOS][1] < 20 and haut_ligne is not None and est_vide(monde, haut_ligne, haut_colonne):
        deplacer_animal(monde, infos, liste_moutons, loup, haut_ligne, haut_colonne)
        loup[INDICE_INFOS][1] += 1
    elif loup[INDICE_INFOS][2] < 20 and droite_ligne is not None and est_vide(monde, droite_ligne, droite_colonne):
        deplacer_animal(monde, infos, liste_moutons, loup, droite_ligne, droite_colonne)
        loup[INDICE_INFOS][2] += 1
    elif loup[INDICE_INFOS][3] < 20 and bas_ligne is not None and est_vide(monde, bas_ligne, bas_colonne):
        deplacer_animal(monde, infos, liste_moutons, loup, bas_ligne, bas_colonne)
        loup[INDICE_INFOS][3] += 1
    else:
        loup[INDICE_INFOS] = [0, 0, 0, 0]


def deplacer_un_loup_encercler(monde, infos, liste_loups, liste_moutons, loup):
    """
    Fonction déplaçant un loup. Les loups se choisissent une cible et essaient de l'encercler.
    ici, le paramètre infos représente la proie courante.
    """
    ligne = loup[INDICE_LIGNE]
    colonne = loup[INDICE_COLONNE]
    if infos is None:
        if len(liste_moutons) == 0:
            return
        num_proie = randint(0, len(liste_moutons) - 1)
        infos = liste_moutons[num_proie]
    ligne_proie_courante = infos[INDICE_LIGNE]
    colonne_proie_courante = infos[INDICE_COLONNE]
    
    destinations_possibles = cases_vers(ligne, colonne, ligne_proie_courante, colonne_proie_courante)
    if len(destinations_possibles) == 0:
        destinations_possibles =  voisins_libres(monde, ligne, colonne)
    if len(destinations_possibles) > 0:
        num_destination = randint(0, len(destinations_possibles) - 1)
        destination = destinations_possibles[num_destination]
        deplacer_animal(monde, infos, liste_moutons, loup, destination[0], destination[1])
    if est_vide(monde, ligne_proie_courante, colonne_proie_courante):
        #la proie a été mangée
        infos = None

def deplacer_un_loup_groupe(monde, infos, liste_loups, liste_moutons, loup):
    """
    Fonction déplaçant un loup. Les loups proches se liguent vers une cible donnée.
    """
    ligne = loup[INDICE_LIGNE]
    colonne = loup[INDICE_COLONNE]
    mouton_cible = loup[INDICE_INFOS]
    if mouton_cible is not None:
        if monde[mouton_cible[INDICE_LIGNE]][mouton_cible[INDICE_COLONNE]] != mouton_cible:
            # le mouton ciblé est mort (de stress) entre temps
            loup[INDICE_INFOS] = None
    if loup[4] is None:
        loups_proches = loups_les_plus_proches(ligne, colonne, TAILLE, liste_loups)
        for autre_loup in loups_proches:
            if autre_loup[INDICE_INFOS] is not None and monde[autre_loup[INDICE_INFOS][INDICE_LIGNE]][autre_loup[INDICE_INFOS][INDICE_COLONNE]] is not None:
                loup[INDICE_INFOS] = autre_loup[INDICE_INFOS]
                break
        if loup[INDICE_INFOS] is None:
            loup[INDICE_INFOS] = mouton_le_plus_proche(ligne, colonne, liste_moutons)
            if loup[INDICE_INFOS] is None:
                return
    proie_courante = loup[INDICE_INFOS]
    destinations_possibles = cases_vers(ligne, colonne, proie_courante[INDICE_LIGNE], proie_courante[INDICE_COLONNE])
    if len(destinations_possibles) == 0:
        destinations_possibles =  voisins_libres(monde, ligne, colonne)
    if len(destinations_possibles) > 0:
        num_destination = randint(0, len(destinations_possibles) - 1)
        destination = destinations_possibles[num_destination]
        deplacer_animal(monde, infos, liste_moutons, loup, destination[0], destination[1])
        if est_vide(monde, proie_courante[INDICE_LIGNE], proie_courante[INDICE_COLONNE]):
            #la proie a été mangée
            for autre_loup in liste_loups:
                if autre_loup[INDICE_INFOS] == loup[INDICE_INFOS]:
                    autre_loup[INDICE_INFOS] = None
            loup[INDICE_INFOS] = None

            
def deplacer_un_loup(monde, infos, liste_loups, liste_moutons, loup):
    """
    Fonction déplaçant un loup. Cette fonction propose un déplacement de base,
    totalement aléatoire. Le but est de modifier cette fonction (en introduisant
    éventuellement de nouvelles sous-fonctions) pour que la capture des loups
    soit plus efficace.
    """
    ligne = loup[INDICE_LIGNE]
    colonne = loup[INDICE_COLONNE]
    cases_possibles =  voisins_libres(monde, ligne, colonne)
    if len(cases_possibles) > 0:
        num_destination = randint(0, len(cases_possibles) - 1)
        destination = cases_possibles[num_destination]
        deplacer_animal(monde, infos, liste_moutons, loup, destination[0], destination[1])

def deplacer_les_loups(monde, infos, liste_loups, liste_moutons):
    """
    Fonction faisant se déplacer tous les loups
    """
    for loup in liste_loups:
        deplacer_un_loup(monde, infos, liste_loups, liste_moutons, loup)

def animer(nb_moutons, nb_loups, affichage=True):
    """
    Fonction permettant de lancer l'animation : on commence par créer
    le monde puis le peupler avec le nombre de moutons et le nombre
    de loups passés en paramètre.
    Ensuite on exécute des "tours" jusqu'à ce que tous les moutons aient
    disparus (attrapés par les loups ou mort de stress).
    À chaque tour, ce sont soit tous les loups (probabilité : 75%) soit
    tous les moutons (probabilité : 25%) qui se déplacent.
    Par défaut, l'animation affiche l'état du monde toutes les 10 étapes.
    Pour aller plus vite ou en cas de problème d'affichage, il est possible
    de supprimer ces afficahge en passant "False" en 3è paramètre lors
    de l'appel de la fonction 'animer'.
    """
    monde = creer_monde()
    infos_generales = None
    moutons, loups = peupler_monde(monde, infos_generales, nb_moutons, nb_loups)
    if affichage:
        dessiner_monde(monde)
    nb_iterations = 0
    while len(moutons) > 0:
        nb_iterations += 1
        p = randint(0, 100)
        if p < 75:
            # 75% de chance d'activer les loups
            deplacer_les_loups(monde, infos_generales, loups, moutons)
        else:
            # dans les 25% restant, on active les moutons
            deplacer_les_moutons(monde, infos_generales, moutons)
        if affichage:
            if nb_iterations % 10 == 0:
                dessiner_monde(monde)
                time.sleep(1)
            else:
                time.sleep(0.2)
    print("Tous les moutons ont été attrapés en", nb_iterations, "coups.")
